import os


class Config(object):
    PROJECT_ID = os.environ.get('GCREDSTASH_GCP_PROJECT_ID')
    DEFAULT_KEY_RING_ID = os.environ.get('GCREDSTASH_DEFAULT_KEY_RING_ID')
    DEFAULT_LOCATION_ID = os.environ.get('GCREDSTASH_DEFAULT_LOCATION_ID') or "global"
    DEFAULT_CRYPTO_KEY_ID = os.environ.get('GCREDSTASH_DEFAULT_CRYPTO_KEY_ID')
    DEFAULT_DATASTORE_KIND = os.environ.get('GCREDSTASH_DEFAULT_DATASTORE_KIND') or "Credentials"



from __future__ import print_function
import argparse
import sys

import googleapiclient.discovery

from config import Config
from keystore import KeyStore
from kms import GoogleKMS

parser = argparse.ArgumentParser(description='A Credential Management Tool using Google Cloud KMS and Datastore')
subparsers = parser.add_subparsers(dest='command',
                                   help='Try commands like "{name} get -h" to get sub command\'s options'.format(
                                       name=sys.argv[0])
                                   )
create_key_ring_parser = subparsers.add_parser('create-keyring',
                                               help="Creates a KeyRing in the given location (e.g. global)")
create_key_ring_parser.add_argument('new_keyring_id', type=str, help='Unique id for new KeyRing creation in location')

create_key_parser = subparsers.add_parser('create-key',
                                          help="Creates a CryptoKey within a KeyRing in the given location")
create_key_parser.add_argument('new_key_id', type=str, help='Unique id for new key creation in the KeyRing')

get_parser = subparsers.add_parser('get', help="Get the cipher from KeyStore and decrypted the credential")
get_parser.add_argument('name', type=str, help='The name of credential')

put_parser = subparsers.add_parser('put', help="Encrypt the credential and put the cipher on KeyStore")
put_parser.add_argument('name', type=str, help='The name of credential')
put_parser.add_argument('plaintext', type=str, help='Text to be encrypted')

get_all_parser = subparsers.add_parser('get-all', help="Get the cipher from KeyStore and decrypted all the credentials")
list_parser = subparsers.add_parser('list')

parser.add_argument(
    '--project-id',
    default=Config.PROJECT_ID, dest='project_id', type=str, help='GCP Project Id')
parser.add_argument(
    '--location-id',
    default=Config.DEFAULT_LOCATION_ID, dest='location_id', type=str, help='Google Cloud KMS Location Id')
parser.add_argument(
    '--keyring-id',
    default=Config.DEFAULT_KEY_RING_ID, dest='key_ring_id', type=str, help='Google Cloud KMS KeyRing Id')
parser.add_argument(
    '--key-id',
    default=Config.DEFAULT_CRYPTO_KEY_ID, dest='key_id', type=str, help='Google Cloud KMS CryptoKey Id')


def main():
    args = parser.parse_args()
    key_store = KeyStore(args.project_id)
    kms_client = googleapiclient.discovery.build('cloudkms', 'v1')
    kms = GoogleKMS(kms_client, args.project_id, args.location_id, args.key_ring_id, key_store)

    if args.command == "put":
        kms.put(args.key_id, Config.DEFAULT_DATASTORE_KIND, args.name, args.plaintext)
    elif args.command == "get":
        print(kms.get(args.key_id, Config.DEFAULT_DATASTORE_KIND, args.name))
    elif args.command == "get-all":
        print(kms.get_all(args.key_id, Config.DEFAULT_DATASTORE_KIND))
    elif args.command == "create-keyring":
        print(kms.create_key_ring(kms_client, args.project_id, args.location_id, args.new_keyring_id))
    elif args.command == "create-key":
        print(kms.create_key(kms_client, args.project_id, args.location_id, args.key_ring_id, args.new_key_id))
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



from __future__ import print_function
import base64
import logging


class GoogleKMS(object):
    PARENT_URI = 'projects/{project_id}/locations/{location_id}'
    KEY_RING_URI = PARENT_URI + '/keyRings/{key_ring_id}'
    KEY_URI = KEY_RING_URI + '/cryptoKeys/{key_id}'

    def __init__(self, client, project_id, location_id, key_ring_id, key_store):
        self.client = client
        self.project_id = project_id
        self.location_id = location_id
        self.key_ring_id = key_ring_id
        self.key_store = key_store

    def _encrypt(self, key_id, plain_text):
        """
        Encrypt text using the CryptoKey
        :param key_id: CryptKey id
        :param plain_text: Plain text for encryption
        :return: Cipher used for encryption
        """
        try:

            name = self._get_key_uri(key_id)

            crypto_keys = self._get_crypto_keys()
            request = crypto_keys.encrypt(
                name=name,
                body={'plaintext': base64.b64encode(plain_text.encode('utf8')).decode('utf8')}
            )

            response = request.execute()
            return response.get('ciphertext').encode('utf8')
        except Exception as e:
            logging.exception(e)

    def _decrypt(self, key_id, cipher):
        """
        Decrypt the credential using cipher text
        :param key_id: CryptKey id
        :param cipher: Cipher text for decrypting the encrypted credential
        :return: Decrypted text
        """
        try:
            name = self._get_key_uri(key_id)
            crypto_keys = self._get_crypto_keys()

            request = crypto_keys.decrypt(
                name=name,
                body={'ciphertext': cipher})
            response = request.execute()

            return base64.b64decode(response.get('plaintext').encode('utf8'))

        except Exception as e:
            logging.exception(e)

    def _get_crypto_keys(self):
        """
        Get Google KMS CryptKeys in the KeyRing
        :return: CryptoKeys
        """
        return self.client.projects().locations().keyRings().cryptoKeys()

    def _get_key_uri(self, key_id):
        """
        Get Google Cloud KMS resource id
        :param key_id: CryptKey id
        :return:
        """
        return GoogleKMS.KEY_URI.format(
            project_id=self.project_id,
            location_id=self.location_id,
            key_ring_id=self.key_ring_id,
            key_id=key_id,

        )

    def get(self, key_id, kind, name):
        """
        Get the cipher from KeyStore and decrypted the credential
        :param key_id: CryptKey id
        :param kind: Keystore collection name
        :param name: Name of the credential to fetch ex: PRODUCTION_DATABASE_PASSWORD
        :return: Decrypted text
        """

        if not isinstance(name, basestring):
            raise ValueError("name and value should be a string")

        cipher = self.key_store.get(kind, name)
        return self._decrypt(key_id, cipher)

    def put(self, key_id, kind, name, value):
        """
        Encrypt the credential and put the cipher on KeyStore
        :param key_id: CryptKey id
        :param kind: Keystore collection name
        :param name: Name of the credential to store ex: PRODUCTION_DATABASE_PASSWORD
        :param value: Value for the credential
        :return:
        """

        if not isinstance(name, basestring) or not isinstance(value, basestring):
            raise ValueError("name and value should be a string")

        cipher = self._encrypt(key_id, value)
        return self.key_store.put(kind, name, cipher)

    def put_all(self, key_id, kind, credentials):
        """
        Encrypt all the credential and store the cipher on KeyStore
        :param key_id: CryptKey id
        :param kind: Keystore collection name
        :param credentials: Credentials {name: value} dictionary

        :return:
        """

        if type(credentials) != dict:
            raise ValueError("credentials should be a dictionary of credential name and value")

        for name, value in credentials.items():
            self.put(key_id, kind, name, value)

    def get_all(self, key_id, kind):
        """
        Get the cipher from KeyStore and decrypted all the credentials
        :param key_id: CryptKey id
        :param kind: Keystore collection name

        :return: Dictionary of credentials
        """

        return {name: self.get(key_id, kind, name) for name in self.key_store.list(kind)}

    @staticmethod
    def create_key_ring(client, project_id, location_id, key_ring_id):
        """
        Creates a KeyRing in the given location (e.g. global)
        :param client: Google Cloud KMS client
        :param project_id: GCP Project id
        :param location_id: KeyRing location id
        :param key_ring_id: New unique key ring name
        :return: KeyRing creation response object
        """
        try:
            key_rings = client.projects().locations().keyRings()
            parent_uri = GoogleKMS.PARENT_URI.format(
                project_id=project_id,
                location_id=location_id
            )

            request = key_rings.create(parent=parent_uri, body={}, keyRingId=key_ring_id)
            return request.execute()
        except Exception as e:
            logging.exception(e)

    @staticmethod
    def create_key(client, project_id, location_id, key_ring_id, key_id):
        """
        Creates a CryptoKey within a KeyRing in the given location
        :param client: Google Cloud KMS client
        :param project_id: GCP Project id
        :param location_id: KeyRing location id
        :param key_ring_id: KeyRing id inside the location
        :param key_id:  New unique key name
        :return: Key creation response object
        """

        try:
            parent_uri = GoogleKMS.KEY_RING_URI.format(
                project_id=project_id,
                location_id=location_id,
                key_ring_id=key_ring_id
            )

            crypto_keys = client.projects().locations().keyRings().cryptoKeys()

            request = crypto_keys.create(
                parent=parent_uri, body={'purpose': 'ENCRYPT_DECRYPT'},
                cryptoKeyId=key_id)
            return request.execute()
        except Exception as e:
            logging.exception(e)



from google.cloud import datastore


class KeyStore(object):
    CIPHER_PROPERTY_KEY = 'cipher'

    def __init__(self, project_id=None, namespace=None):
        self.client = datastore.Client(project=project_id, namespace=namespace)

    def get(self, kind, name):
        """
        Get value associated with the name from Datastore
        :param kind: Collection name
        :param name: Datastore key
        :return: str
        """
        key = self.client.key(kind, name)

        entity = self.client.get(key)
        if not entity:
            return None

        return entity.get(KeyStore.CIPHER_PROPERTY_KEY)

    def put(self, kind, name, content):
        """
        Put value on the Datastore
        :param kind: Collection name
        :param name: Datastore key
        :param content: value to store
        :return:
        """
        key = self.client.key(kind, name)
        entity = datastore.Entity(key=key, exclude_from_indexes=(KeyStore.CIPHER_PROPERTY_KEY,))
        entity[KeyStore.CIPHER_PROPERTY_KEY] = content
        return self.client.put(entity)

    def list(self, kind):
        """

        :param kind: Collection name
        :return: list of Datastore keys
        """

        return [c.key.name for c in self.client.query(kind=kind).fetch()]



from setuptools import setup

setup(
    name='gcredstash',
    author='Rajesh Hegde',
    version='1.0.1',
    description='A Credential Management Tool using Google Cloud KMS and Datastore',
    url='https://github.com/RajeshHegde/gcredstash',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
    ],
    packages=['gcredstash'],
    install_requires=[
        'google-api-python-client>=1.6.5',
        'google-cloud-datastore>=1.6.0',
        'google-auth-httplib2>=0.0.3',
    ],
    entry_points={
        'console_scripts': [
            'gcredstash = gcredstash.main:main'
        ]
    }
)
